function [D, X, iter] = indef_lobpcg_wh(A, B, funC, X, tol)

[n, nev] = size(X);
D = zeros(nev, 1);
normA = normest(A, 0.1);
normB = normest(B, 0.1);

% Initialization
AX = A*X; BX = B*X;
[X, AX, BX, d, m] = MGS2(X, AX, BX, A, B); % Bp=X'*BX=diag(d)
Ap = X'*AX;
Bp = X'*BX;
Ap = (Ap+Ap') / 2; Bp = (Bp+Bp') / 2;
[V0, D0] = eig(Ap, Bp);
D0 = diag(D0);
D0 = real(D0)+abs(imag(D0));
[D0, ind] = sort(D0);
V0 = real(V0(:, ind));
X = X*V0;
AX = AX*V0; BX=BX*V0;
n1 = sum(d>=0);
n2 = sum(d<0);
% This algorithm finds n1 smallest eigenvalues with positive sign
% and n2 largest eigenvalues with negative sign, i.e. the inertia
% of X'BX keeps unchanged.

deflated = 0;
deflated_pos = 0;
deflated_neg = 0;
indX = 1 : nev;
PH = [];
APH = [];
BPH = [];
gnorm = 1;
iter=0;
while gnorm > tol
    iter=iter+1;
     R = AX - D0*BX;
    gnorm=norm(R,2);
    indX = 1 : nev-deflated;
    active = deflated + indX;

    % Generate projection subspace
    W=funC(R(:, active));
    AW = A*W; BW = B*W;
    if iter == 1,
        V = [X, W, PH];
        sv=size(V);
        m = sv(2);
        AV = [AX, AW, APH]; BV = [BX, BW, BPH];
%        m = 2*(nev-deflated);
%        indP = nev + active;
    else
        V = [X, W, PH];
        sv=size(V);
        m = sv(2);
        AV = [AX, AW, APH]; BV = [BX, BW, BPH];
%        m = 3*(nev-deflated);
%        indP = [nev + active, 2*nev-deflated + active];
    end
    [V, AV, BV, d, m] = MGS2(V, AV, BV, A, B);
    while m <= n1+n2,
        V = [X, randn(size([P(:, active), W]))];
        [V, AV, BV, d, m] = MGS2(V, AV, BV, A, B);
    end
    m1 = sum(d>=0);
    m2 = sum(d<0);
    if(m1 < n1 || m2 < n2),
        error('This will never happen in exact arithmetic.');
    end

    % Exclude converged eigenvectors
    V = V(:, deflated+1 : end);
    AV = AV(:, deflated+1 : end);
    BV = BV(:, deflated+1 : end);
    m = size(V, 2);
    % Solve projected problem
    Ap = V'*AV;
    Bp = V'*BV; % In fact, Bp = diag {\pm 1};
    Ap = (Ap+Ap') / 2; Bp = (Bp+Bp') / 2;
    [V0, D0] = eig(Ap, Bp);
    D0 = diag(D0);
    D0 = real(D0)+imag(D0);
    % Extract n1-deflated_pos smallest ``positive'' eigenvalues
    % and n2-deflated_neg largest ``negative'' eigenvalues
    [D0, ind] = sort(D0);
    V0 = real(V0(:, ind));
    D0(active) =D0(m2-n2 + indX);
    D0=D0(1);
    V0 = V0(:, m2-n2 + indX);%V0(:,1);
    indP = nev-deflated+1 : m;

    % Update P and X
    
         P(:, active) = V(:, indP) * V0(indP, :);
         AP(:, active) = AV(:, indP) * V0(indP, :);
         BP(:, active) = BV(:, indP) * V0(indP, :);
         X(:, active) =  P(:, active) + V(:, indX) * V0(indX, :);
         AX(:, active) = AP(:, active) + AV(:, indX) * V0(indX, :);
         BX(:, active) = BP(:, active) + BV(:, indX) * V0(indX, :);
        
    PH = [P(:, active) PH];
    APH = [AP(:, active) APH];
    BPH = [BP(:, active) BPH];
%    PH = P(:, active);
%    APH = AP(:, active);
%    BPH = BP(:, active);
end
% for iter = 1 : maxiter
%     % Check converged eigenvalues
%     normD = norm(D);
%     maxr = 0;%
%     j = deflated + n2 - deflated_neg;
%     while j > deflated
%         R(:, j) = AX(:, j) - D0(j)*BX(:, j);
% %        r = norm(R(:, j), 2) / (norm(AX(:, j)) + abs(D0(j))*norm(BX(:, j)));
%         r = norm(R(:, j), 2) / (normA + abs(D0(j))*normB);
%         if r > maxr, maxr = r; end %
%         % deflate ``negative'' eigenvalues
%         if iter > 1 && maxr <= tol && deflated_neg < n2,
%             deflated_neg = deflated_neg+1;
%             deflated = deflated+1;
%             D(deflated) = D0(j);
%             if (j > deflated),
%                  X(:, [deflated, deflated+1:j]) =  X(:, [j, deflated:j-1]);
%                 AX(:, [deflated, deflated+1:j]) = AX(:, [j, deflated:j-1]);
%                 BX(:, [deflated, deflated+1:j]) = BX(:, [j, deflated:j-1]);
%                  P(:, [deflated, deflated+1:j]) =  P(:, [j, deflated:j-1]);
%                 AP(:, [deflated, deflated+1:j]) = AP(:, [j, deflated:j-1]);
%                 BP(:, [deflated, deflated+1:j]) = BP(:, [j, deflated:j-1]);
%                  R(:, [deflated, deflated+1:j]) =  R(:, [j, deflated:j-1]);
%                    D0([deflated, deflated+1:j]) =    D0([j, deflated:j-1]);
%             end
%         else
%             j = j - 1;
%         end
%     end
%     maxr = 0;
%     for j = deflated + n2 - deflated_neg + 1 : nev
%         R(:, j) = AX(:, j) - D0(j)*BX(:, j);
% %        r = norm(R(:, j), 2) / (norm(AX(:, j)) + abs(D0(j))*norm(BX(:, j)));
%         r = norm(R(:, j), 2) / (normA + abs(D0(j))*normB);
%         if r > maxr, maxr = r; end %
%         % deflate ``positive'' eigenvalues
%         if iter > 1 && maxr <= tol && deflated_pos < n1,
%             deflated_pos = deflated_pos+1;
%             deflated = deflated+1;
%             D(deflated) = D0(j);
%             if(j > deflated),
%                 X(:, [deflated, j]) = X(:, [j, deflated]);
%                 AX(:, [deflated, j]) = AX(:, [j, deflated]);
%                 BX(:, [deflated, j]) = BX(:, [j, deflated]);
%                 P(:, [deflated, j]) = P(:, [j, deflated]);
%                 AP(:, [deflated, j]) = AP(:, [j, deflated]);
%                 BP(:, [deflated, j]) = BP(:, [j, deflated]);
%                 R(:, [deflated, j]) = R(:, [j, deflated]);
%             end
%         end
%       end
% %     fprintf('step %3d,  deflated %2d (%2d+%2d),  r = %.2e \n',...
% %         iter, deflated, deflated_neg, deflated_pos, maxr);
%     if deflated >= nev, break; end
%     indX = 1 : nev-deflated;
%     active = deflated + indX;
% 
%     % Generate projection subspace
%     W = T*R(:, active);
%     AW = A*W; BW = B*W;
%     if iter == 1,
%         V = [X, W, PH];
%         sv=size(V);
%         m = sv(2);
%         AV = [AX, AW, APH]; BV = [BX, BW, BPH];
% %        m = 2*(nev-deflated);
% %        indP = nev + active;
%     else,
%         V = [X, W, PH];
%         sv=size(V);
%         m = sv(2);
%         AV = [AX, AW, APH]; BV = [BX, BW, BPH];
% %        m = 3*(nev-deflated);
% %        indP = [nev + active, 2*nev-deflated + active];
%     end
%     [V, AV, BV, d, m] = MGS2(V, AV, BV, A, B);
%     while m <= n1+n2,
%         V = [X, randn(size([P(:, active), W]))];
%         [V, AV, BV, d, m] = MGS2(V, AV, BV, A, B);
%     end
%     m1 = sum(d>=0);
%     m2 = sum(d<0);
%     if(m1 < n1 || m2 < n2),
%         error('This will never happen in exact arithmetic.');
%     end
% 
%     % Exclude converged eigenvectors
%     V = V(:, deflated+1 : end);
%     AV = AV(:, deflated+1 : end);
%     BV = BV(:, deflated+1 : end);
%     m = size(V, 2);
%     % Solve projected problem
%     Ap = V'*AV;
%     Bp = V'*BV; % In fact, Bp = diag {\pm 1};
%     Ap = (Ap+Ap') / 2; Bp = (Bp+Bp') / 2;
%     [V0, D0] = eig(Ap, Bp);
%     D0 = diag(D0);
%     D0 = real(D0)+imag(D0);
%     % Extract n1-deflated_pos smallest ``positive'' eigenvalues
%     % and n2-deflated_neg largest ``negative'' eigenvalues
%     [D0, ind] = sort(D0);
%     V0 = real(V0(:, ind));
%     D0(active) = D0(m2-n2 + indX);
%     V0 = V0(:, m2-n2 + indX);
%     indP = nev-deflated+1 : m;
% 
%     % Update P and X
%     
%          P(:, active) = V(:, indP) * V0(indP, :);
%          sp=size(P);
%          AP(:, active) = AV(:, indP) * V0(indP, :);
%          BP(:, active) = BV(:, indP) * V0(indP, :);
%          X(:, active) =  P(:, active) + V(:, indX) * V0(indX, :);
%          AX(:, active) = AP(:, active) + AV(:, indX) * V0(indX, :);
%          BX(:, active) = BP(:, active) + BV(:, indX) * V0(indX, :);
%     
% %     PH = [P(:, active) PH];
% %     APH = [AP(:, active) APH];
% %     BPH = [BP(:, active) BPH];
%    PH = [P(:, active)];
%    APH = [AP(:, active)];
%    BPH = [BP(:, active)];
% end
if deflated < nev,
    D(deflated+1:end)=D0(active);
end
[D, ind] = sort(D);
X = X(:, ind);
return
